@@ -368,30 +368,21 @@ const (
 	C_LCON   /* other 32 */
 	C_DCON   /* other 64 (could subdivide further) */
 	C_SACON  /* $n(REG) where n <= int16 */
-	C_SECON
-	C_LACON /* $n(REG) where int16 < n <= int32 */
-	C_LECON
-	C_DACON /* $n(REG) where int32 < n */
+	C_LACON  /* $n(REG) where int16 < n <= int32 */
+	C_DACON  /* $n(REG) where int32 < n */
 	C_SBRA
 	C_LBRA
 	C_LBRAPIC
-	C_SAUTO
-	C_LAUTO
-	C_SEXT
-	C_LEXT
 	C_ZOREG // conjecture: either (1) register + zeroed offset, or (2) "R0" implies zero or C_REG
-	C_SOREG // register + signed offset
-	C_LOREG
+	C_SOREG // D/DS form memory operation
+	C_LOREG // 32 bit addis + D/DS-form memory operation
 	C_FPSCR
-	C_MSR
 	C_XER
 	C_LR
 	C_CTR
 	C_ANY
 	C_GOK
 	C_ADDR
-	C_GOTADDR
-	C_TOCADDR
 	C_TLS_LE
 	C_TLS_IE
 	C_TEXTSIZE
@@ -458,6 +449,10 @@ const (
 	ADIVWUCC
 	ADIVWUVCC
 	ADIVWUV
+	AMODUD
+	AMODUW
+	AMODSD
+	AMODSW
 	AEQV
 	AEQVCC
 	AEXTSB
@@ -565,18 +560,13 @@ const (
 	AORNCC
 	AORIS
 	AREM
-	AREMCC
-	AREMV
-	AREMVCC
 	AREMU
-	AREMUCC
-	AREMUV
-	AREMUVCC
 	ARFI
 	ARLWMI
 	ARLWMICC
 	ARLWNM
 	ARLWNMCC
+	ACLRLSLWI
 	ASLW
 	ASLWCC
 	ASRW
@@ -584,6 +574,7 @@ const (
 	ASRAWCC
 	ASRWCC
 	ASTBCCC
+	ASTHCCC
 	ASTSW
 	ASTWCCC
 	ASUB
@@ -717,6 +708,9 @@ const (
 	ARLDCLCC
 	ARLDICL
 	ARLDICLCC
+	ARLDIC
+	ARLDICCC
+	ACLRLSLDI
 	AROTL
 	AROTLW
 	ASLBIA
@@ -730,25 +724,25 @@ const (
 	ASRAD
 	ASRADCC
 	ASRDCC
+	AEXTSWSLI
+	AEXTSWSLICC
 	ASTDCCC
 	ATD
 
 	/* 64-bit pseudo operation */
 	ADWORD
 	AREMD
-	AREMDCC
-	AREMDV
-	AREMDVCC
 	AREMDU
-	AREMDUCC
-	AREMDUV
-	AREMDUVCC
 
 	/* more 64-bit operations */
 	AHRFID
 	APOPCNTD
 	APOPCNTW
 	APOPCNTB
+	ACNTTZW
+	ACNTTZWCC
+	ACNTTZD
+	ACNTTZDCC
 	ACOPY
 	APASTECC
 	ADARN
@@ -902,8 +896,15 @@ const (
 	AVCMPGTSDCC
 	AVCMPNEZB
 	AVCMPNEZBCC
+	AVCMPNEB
+	AVCMPNEBCC
+	AVCMPNEH
+	AVCMPNEHCC
+	AVCMPNEW
+	AVCMPNEWCC
 	AVPERM
 	AVPERMXOR
+	AVPERMR
 	AVBPERMQ
 	AVBPERMD
 	AVSEL
@@ -925,31 +926,37 @@ const (
 	AVSHASIGMA
 	AVSHASIGMAW
 	AVSHASIGMAD
+	AVMRGEW
+	AVMRGOW
 
 	/* VSX */
 	ALXV
+	ALXVL
+	ALXVLL
 	ALXVD2X
-	ALXVDSX
 	ALXVW4X
+	ALXVH8X
+	ALXVB16X
+	ALXVX
+	ALXVDSX
 	ASTXV
+	ASTXVL
+	ASTXVLL
 	ASTXVD2X
 	ASTXVW4X
-	ALXS
+	ASTXVH8X
+	ASTXVB16X
+	ASTXVX
 	ALXSDX
-	ASTXS
 	ASTXSDX
-	ALXSI
 	ALXSIWAX
 	ALXSIWZX
-	ASTXSI
 	ASTXSIWX
-	AMFVSR
 	AMFVSRD
 	AMFFPRD
 	AMFVRD
 	AMFVSRWZ
 	AMFVSRLD
-	AMTVSR
 	AMTVSRD
 	AMTFPRD
 	AMTVRD
@@ -958,7 +965,6 @@ const (
 	AMTVSRDD
 	AMTVSRWS
 	AXXLAND
-	AXXLANDQ
 	AXXLANDC
 	AXXLEQV
 	AXXLNAND
@@ -968,34 +974,32 @@ const (
 	AXXLORQ
 	AXXLXOR
 	AXXSEL
-	AXXMRG
 	AXXMRGHW
 	AXXMRGLW
 	AXXSPLT
 	AXXSPLTW
+	AXXSPLTIB
 	AXXPERM
 	AXXPERMDI
-	AXXSI
 	AXXSLDWI
-	AXSCV
+	AXXBRQ
+	AXXBRD
+	AXXBRW
+	AXXBRH
 	AXSCVDPSP
 	AXSCVSPDP
 	AXSCVDPSPN
 	AXSCVSPDPN
-	AXVCV
 	AXVCVDPSP
 	AXVCVSPDP
-	AXSCVX
 	AXSCVDPSXDS
 	AXSCVDPSXWS
 	AXSCVDPUXDS
 	AXSCVDPUXWS
-	AXSCVXP
 	AXSCVSXDDP
 	AXSCVUXDDP
 	AXSCVSXDSP
 	AXSCVUXDSP
-	AXVCVX
 	AXVCVDPSXDS
 	AXVCVDPSXWS
 	AXVCVDPUXDS
@@ -1004,7 +1008,6 @@ const (
 	AXVCVSPSXWS
 	AXVCVSPUXDS
 	AXVCVSPUXWS
-	AXVCVXP
 	AXVCVSXDDP
 	AXVCVSXWDP
 	AXVCVUXDDP
